'''
Facilitate the packaging and deployment of the function
'''
# pylint: disable=invalid-name
# pylint: disable=broad-except
# pylint: disable=line-too-long
# pylint: disable=bare-except

import logging
import traceback
import subprocess
import os
import git
import sys
import uuid
import json
import boto3
import shutil
import zipfile

from configparser import ConfigParser
from lambdatool.stack_tool import StackTool
from lambdatool.template_creator import TemplateCreator
from stackility import CloudStackUtility
from lambdatool.pip_util import pipmain

try:
    import zlib #noqa
    compression = zipfile.ZIP_DEFLATED
except:
    compression = zipfile.ZIP_STORED

logging.basicConfig(level=logging.INFO,
                    format='[%(levelname)s] %(asctime)s (%(module)s) %(message)s',
                    datefmt='%Y/%m/%d-%H:%M:%S')

logging.getLogger().setLevel(logging.INFO)


LAMBDATOOL_VERSION = '0.9.3'
LAMBDATOOL_DESCRIPTOR = '.lambdatool'
DEFAULT_DESCRIPTION = 'Fantastic Lambda Function'
DEFAULT_MODULE_FILE = 'main.py'
IGNORED_STUFF = ('config', '.git')
PIP_ARGS = [
    'install',
    '-Ur',
    'requirements.txt',
    '-t',
    '.'
]

ZIP_MODES = {
    zipfile.ZIP_DEFLATED: 'deflated',
    zipfile.ZIP_STORED:   'stored'
}


class LambdaDeployer:
    """
    Lambda utility is yet another tool create and deploy AWS lambda functions
    """
    _ini_data = None
    _stage = None
    _work_directory = None
    _package_name = None
    _package_key = None
    _profile = None
    _region = None
    _lambda_name = None
    _create_log_group = False
    _description = None
    _hash = None
    _tag_file = None
    _stack_properties = {}
    _template_directory = None
    _s3_client = None
    _apig_client = None
    _ssm_client = None
    _python = None

    def __init__(self, config_block):
        """
        Lambda deployer init method.

        Args:
            config_block - a dictionary created in the CLI driver. See that
                           script for the things that are required and
                           optional.

        Returns:
           not a damn thing

        Raises:
            SystemError - if everything isn't just right
        """
        self._stack_name = None
        if config_block:
            if not os.path.isdir(config_block['work_directory']):
                logging.error('{} does not exist'.format(config_block['work_directory']))
                raise SystemError

            tmp_name = (str(uuid.uuid4()))[:8]
            self._work_directory = '{}/{}'.format(config_block['work_directory'], tmp_name)
            self._stage = config_block['stage']
            self._profile = config_block['profile']
            self._region = config_block['region']
            self._template_directory = config_block['template_directory']
            self._init_boto3_clients()
        else:
            logging.error('config block was garbage')
            raise SystemError

        v = sys.version_info
        if v.major == 2:
            self._python = 'python2.7'
        elif v.major == 3:
            if v.minor == 6:
                self._python = 'python3.6'
            elif v.minor == 7:
                self._python = 'python3.7'
            elif v.minor == 8:
                self._python = 'python3.8'
            elif v.minor == 9:
                self._python = 'python3.9'
            else:
                logging.error('python %s.%s detected', v.major, v.minor)
                logging.error('only python3.6, python3.7, python3.8 and python3.9 are available')
                raise SystemError
        else:
            logging.error('strange python version')
            raise SystemError
        logging.info('{} runtime selected'.format(self._python))

    def deploy_lambda(self):
        """
        Deploy an existing lambda to the indicated by creating CF template...

        Args:
            None

        Returns:
            True if the lambda is deployed
            False if the lambda is not deployed for some odd reason
        """
        try:
            if not self.read_descriptor():
                logging.error('problem reading ".lambdatool" descriptor, exiting')
                return False

            if not self.verify_lambda_directory():
                logging.error('module file {} not found, exiting'.format(DEFAULT_MODULE_FILE))
                return False

            if self.read_config_info():
                logging.info('INI stuff: {}'.format(json.dumps(self._ini_data, indent=2)))
            else:
                logging.error('failed to read config/config.ini file, exiting'.format(DEFAULT_MODULE_FILE))
                return False

            if self.set_hash():
                logging.info('deploying version/commit {} of {}'.format(self._hash, self._lambda_name))
            else:
                logging.error('git repo required, exiting!')
                return False

            if self.set_package_name():
                logging.info('package_name: {}'.format(self._package_name))
            else:
                logging.error('failed to set package_name')
                return False

            if self.copy_stuff():
                logging.info('copy_stuff() to scratch directory successful')
                os.chdir(self._work_directory)
            else:
                logging.error('copy_stuff() to scratch directory failed')
                return False

            if self.install_requirements():
                logging.info('install_requirements() to scratch directory successful')
            else:
                logging.error('install_requirements() to scratch directory failed')
                return False

            if self.create_zip():
                logging.info('create_zip() created {}'.format(self._package_name))
            else:
                logging.info('create_zip() failed to create {}'.format(self._package_name))
                return False

            if self.upload_package():
                logging.info('upload_package() uploaded {}'.format(self._package_name))
            else:
                logging.info('upload_package() failed to upload {}'.format(self._package_name))
                return False

            if self.create_tag_file():
                logging.info('create_tag_file() created')
            else:
                logging.info('create_tag_file() failed')
                return False

            if self.create_stack_properties():
                logging.info('create_stack_properties() created')
            else:
                logging.info('create_stack_properties() failed')
                return False

            if self.create_template_file():
                logging.info('create_template_file() created')
            else:
                logging.info('create_template_file() failed')
                return False

            if self.create_stack():
                logging.info('create_stack() created')
            else:
                logging.info('create_stack() failed')
                return False

            if self.redeploy_api():
                logging.info('redeploy_api() successful')
            else:
                logging.info('redeploy_api() failed')
                return False

            return True
        except Exception as x:
            logging.error('Exception caught in deploy_lambda(): {}'.format(x))
            traceback.print_exc(file=sys.stdout)
            return False

    def redeploy_api(self):
        try:
            rest_api_id = None
            deployment_id = None

            response = self._cf_client.describe_stack_resources(
                StackName=self._stack_name
            )

            for resource in response['StackResources']:
                if resource['ResourceType'] == 'AWS::ApiGateway::RestApi':
                    rest_api_id = resource['PhysicalResourceId']
                elif resource['ResourceType'] == 'AWS::ApiGateway::Deployment':
                    deployment_id = resource['PhysicalResourceId']

            if rest_api_id and deployment_id:
                logging.debug('redeploying the API')
                r = self._apig_client.create_deployment(
                    restApiId=rest_api_id,
                    stageName=self._stage
                )
                logging.debug(r)

            return True
        except Exception as wtf:
            logging.error(wtf, exc_info=True)

        return False

    def create_stack(self):
        try:
            ini_data = {}
            ini_data['environment'] = {}
            ini_data['tags'] = {}
            ini_data['parameters'] = {}

            self._stack_name = '{}-lambda-{}'.format(
                self._stage,
                self._lambda_name
            )

            tmp_env = self._ini_data[self._stage]
            ini_data['environment']['template'] = '{}/template.yaml'.format(self._work_directory)
            ini_data['environment']['bucket'] = tmp_env['bucket']
            ini_data['environment']['stack_name'] = self._stack_name
            ini_data['codeVersion'] = self._hash
            if self._region:
                ini_data['environment']['region'] = self._region

            if self._profile:
                ini_data['environment']['profile'] = self._profile

            ini_data['tags'] = {'tool': 'lambdatool'}
            ini_data['parameters'] = self._stack_properties
            ini_data['yaml'] = True

            stack_driver = CloudStackUtility(ini_data)
            if stack_driver.upsert():
                logging.info('stack create/update was started successfully.')
                if stack_driver.poll_stack():
                    logging.info('stack create/update was finished successfully.')
                    st = StackTool(
                        self._stack_name,
                        self._stage,
                        self._profile,
                        self._region,
                        self._cf_client,
                    )
                    st.print_stack_info()
                    return True
                else:
                    logging.error('stack create/update was did not go well.')
                    return False
            else:
                logging.error('start of stack create/update did not go well.')
                return False
        except Exception as wtf:
            logging.error('Exception caught in create_stack(): {}'.format(wtf))
            traceback.print_exc(file=sys.stdout)
            return False

    def create_template_file(self):
        try:
            function_properties = '{}/config/{}/function.properties'.format(
                self._work_directory,
                self._stage
            )
            output_file = '{}/template.yaml'.format(self._work_directory)
            template_file = '{}/template_template'.format(self._template_directory)
            templateCreator = TemplateCreator(self._ssm_client)

            description = self._ini_data.get(self._stage, {}).get(
                'description',
                DEFAULT_DESCRIPTION
            )

            template_created = templateCreator.create_template(
                function_properties=function_properties,
                description=description,
                stack_properties=self._stack_properties,
                output_file=output_file,
                template_file=template_file,
                region=self._region,
                stage_name=self._stage,
                short_name=self._lambda_name,
                create_log_group=self._create_log_group,
                account=boto3.client('sts').get_caller_identity()['Account']
            )

            return template_created
        except Exception as wtf:
            logging.error('Exception caught in create_template_file(): {}'.format(wtf))
            traceback.print_exc(file=sys.stdout)
            return False

    def create_stack_properties(self):
        try:
            bucket = self._ini_data.get(self._stage, {}).get('bucket', None)
            memory_size = self._ini_data.get(self._stage, {}).get('memory', '128')
            role = self._ini_data.get(self._stage, {}).get('role', None)
            timeout = self._ini_data.get(self._stage, {}).get('timeout', '60')
            security_group = self._ini_data.get(self._stage, {}).get('security_group', None)
            subnets = self._ini_data.get(self._stage, {}).get('subnets', None)
            sns_topic_arn = self._ini_data.get(self._stage, {}).get('snstopicarn', None)
            trusted_service = self._ini_data.get(self._stage, {}).get('trustedservice', None)
            trusted_account = self._ini_data.get(self._stage, {}).get('trustedaccount', None)
            reserved_concurrency = self._ini_data.get(self._stage, {}).get('reservedconcurrency', None)
            lambda_schedule_expression = self._ini_data.get(self._stage, {}).get('scheduleexpression', None)
            service = self._ini_data.get(self._stage, {}).get('service', None)
            export_name = self._ini_data.get(self._stage, {}).get('export_name', None)
            retention_days = self._ini_data.get(self._stage, {}).get('retention_days', '30')
            whitelist = self._ini_data.get(self._stage, {}).get('whitelist', None)

            wrk = {}
            wrk['s3Bucket'] = bucket
            wrk['s3Key'] = self._package_key
            wrk['functionName'] = '{}-{}'.format(self._lambda_name, self._stage)
            wrk['logGroupName'] = '/aws/lambda/{}-{}'.format(self._lambda_name, self._stage)
            wrk['handler'] = 'main.lambda_handler'
            wrk['runTime'] = self._python
            wrk['memorySize'] = memory_size
            wrk['role'] = role
            wrk['timeOut'] = timeout
            wrk['securityGroupIds'] = security_group
            wrk['subnetIds'] = subnets
            wrk['retentionDays'] = retention_days

            if whitelist:
                wrk['whitelist'] = whitelist

            if export_name:
                wrk['export_name'] = export_name

            if sns_topic_arn:
                logging.info('subscribing lambda to SNS topic: {}'.format(sns_topic_arn))
                wrk['snsTopicARN'] = sns_topic_arn

            if trusted_service:
                logging.info('the lambda will trust: {}'.format(trusted_service))
                wrk['trustedService'] = trusted_service

            if trusted_account:
                logging.info('the lambda will trust: {}'.format(trusted_account))
                wrk['trustedAccount'] = trusted_account

            if reserved_concurrency:
                logging.info('the lambda  will reserve {} invocations'.format(reserved_concurrency))
                wrk['reservedConcurrency'] = reserved_concurrency

            if lambda_schedule_expression:
                logging.info('the lambda will be scheduled by: {}'.format(lambda_schedule_expression))
                wrk['scheduleExpression'] = lambda_schedule_expression

            if service:
                if service.lower() == 'true':
                    logging.info('the lambda will integrated with an API gateway')
                    wrk['service'] = 'true'

            self._stack_properties = wrk

            return True
        except Exception as x:
            logging.error('Exception caught in create_stack_properties(): {}'.format(x))
            traceback.print_exc(file=sys.stdout)
            return False

    def create_tag_file(self):
        try:
            self._tag_file = '{}/tag.properties'.format(
                self._work_directory
            )

            logging.info('creating tags file: {}'.format(self._tag_file))
            with open(self._tag_file, "w") as outfile:
                outfile.write('APPLICATION={} lambda function\n'.format(self._lambda_name))
                outfile.write('ENVIRONMENT={}\n'.format(self._stage))
                outfile.write('STACK_NAME=lambda-{}-{}\n'.format(
                    self._lambda_name,
                    self._stage
                ))
                outfile.write('VERSION={}\n'.format(self._hash))
            return True
        except Exception as x:
            logging.error('Exception caught in create_tag_file(): {}'.format(x))
            traceback.print_exc(file=sys.stdout)
            return False

    def upload_package(self):
        try:
            self._package_key = 'lambda-code/{}/{}.zip'.format(
                self._lambda_name,
                self._hash
            )
            logging.info('package key: {}'.format(self._package_key))

            if not self._region:
                self._region = boto3.session.Session().region_name

            bucket = self._ini_data.get(self._stage, {}).get('bucket', None)
            if self._s3_client:
                logging.info('S3 client allocated')
                logging.info('preparing upload to s3://{}/{}'.format(bucket, self._package_key))
            else:
                logging.error('S3 client allocation failed')
                return False

            if bucket:
                with open(self._package_name, 'rb') as the_package:
                    self._s3_client.upload_fileobj(the_package, bucket, self._package_key)
            else:
                logging.error('S3 bucket not found in config/config.ini')
                return False

            return True
        except Exception as wtf:
            logging.error('Exception caught in upload_package(): {}'.format(wtf))
            traceback.print_exc(file=sys.stdout)
            return False

    def install_requirements(self):
        try:
            pipmain(PIP_ARGS)
        except Exception as wtf:
            logging.error('Exception caught in upload_package(): {}'.format(wtf))
            traceback.print_exc(file=sys.stdout)
            return False

        return True

    def copy_stuff(self):
        try:
            shutil.copytree(
                '.',
                self._work_directory,
                ignore=shutil.ignore_patterns(*IGNORED_STUFF)
            )

            shutil.copytree(
                'config/{}'.format(self._stage),
                '{}/config/{}'.format(
                    self._work_directory,
                    self._stage
                )
            )

            return True
        except Exception as miserable_failure:
            logging.error('Exception caught in make_work_dir(): {}'.format(miserable_failure))
            traceback.print_exc(file=sys.stdout)
            return False

    def set_hash(self):
        random_bits = []
        random_bits.append((str(uuid.uuid4()))[:8])
        random_bits.append((str(uuid.uuid4()))[:8])

        try:
            repo = git.Repo(search_parent_directories=False)
            hash = repo.head.object.hexsha
            hash = hash[:8]
        except Exception:
            hash = random_bits[1]

        self._hash = '{}-{}'.format(hash, random_bits[0])
        return True

    def verify_lambda_directory(self):
        return os.path.isfile(DEFAULT_MODULE_FILE)

    def read_descriptor(self):
        try:
            with open(LAMBDATOOL_DESCRIPTOR, 'r') as j:
                stuff = json.load(j)
                self._lambda_name = stuff.get('name', None)
                self._create_log_group = stuff.get('createLogGroup', False)

            if not self._lambda_name:
                cwd = os.getcwd()
                dirs = cwd.split('/')
                self._lambda_name = dirs[-1]

            logging.info('lambda_name: {}'.format(self._lambda_name))
            logging.info('create_log_group: {}'.format(self._create_log_group))
            return True
        except Exception as x:
            logging.error('Exception caught in read_descriptor(): {}'.format(x))
            traceback.print_exc(file=sys.stdout)

        return False

    def execute_command(self, command):
        stdout_buf = str()
        stderr_buf = str()
        try:
            p = subprocess.Popen(command, stdout=subprocess.PIPE)
            out, err = p.communicate()

            if out:
                for c in out:
                    stdout_buf = stdout_buf + c

            if err:
                for c in err:
                    stderr_buf = stderr_buf + c

            return p.returncode, stdout_buf, stderr_buf
        except subprocess.CalledProcessError as x:
            logging.error('Exception caught in execute_command(): {}'.format(x))
            traceback.print_exc(file=sys.stdout)
            return x.returncode, None, None

    def find_data(self, the_dir):
        tree = []
        package_file_name = (self._package_name.split('/'))[-1]
        try:
            for folder, subs, files in os.walk(the_dir):
                for file in files:
                    if file != package_file_name:
                        tree.append('{}/{}'.format(folder, file))
        except Exception:
            pass

        return tree

    def set_package_name(self):
        try:
            if self._work_directory and self._hash:
                self._package_name = '{}/{}.zip'.format(
                    self._work_directory,
                    self._hash
                )
                return True
            else:
                return False
        except Exception:
            return False

    def create_zip(self):
        zf = None
        try:
            logging.info('adding files with compression mode={}'.format(ZIP_MODES[compression]))
            zf = zipfile.ZipFile(self._package_name, mode='w')

            for f in self.find_data('.'):
                zf.write(f, compress_type=compression)

            return True
        except Exception as x:
            logging.error('Exception caught in create_zip(): {}'.format(x))
            traceback.print_exc(file=sys.stdout)
            return False
        finally:
            try:
                zf.close()
            except Exception:
                pass

    def read_config_info(self):
        try:
            ini_file = 'config/config.ini'
            config = ConfigParser()
            config.read(ini_file)
            the_stuff = {}
            for section in config.sections():
                the_stuff[section] = {}
                for option in config.options(section):
                    the_stuff[section][option] = config.get(section, option)

            self._ini_data = the_stuff
            return the_stuff
        except Exception as wtf:
            logging.error('Exception caught in read_config_info(): {}'.format(wtf))
            traceback.print_exc(file=sys.stdout)
            return None

    def _init_boto3_clients(self):
        """
        The utililty requires boto3 clients to SSM, Cloud Formation and S3. Here is
        where we make them.

        Args:
            None

        Returns:
            Good or Bad; True or False
        """
        try:
            if self._profile:
                self._b3Sess = boto3.session.Session(profile_name=self._profile)
            else:
                self._b3Sess = boto3.session.Session()

            self._s3_client = self._b3Sess.client('s3')
            self._cf_client = self._b3Sess.client('cloudformation', region_name=self._region)
            self._ssm_client = self._b3Sess.client('ssm', region_name=self._region)
            self._apig_client = self._b3Sess.client('apigateway', region_name=self._region)

            return True
        except Exception as wtf:
            logging.error('Exception caught in intialize_session(): {}'.format(wtf))
            traceback.print_exc(file=sys.stdout)
            return False



'''
Utililty to find pip's main function
'''
# pylint: disable=invalid-name
# pylint: disable=bare-except

import logging

logger = logging.getLogger(__name__)
if __name__ == '__main__':
    logging.basicConfig(
        level=logging.INFO,
        format='[%(levelname)s] %(asctime)s (%(module)s) %(message)s',
        datefmt='%Y/%m/%d-%H:%M:%S'
    )

pipmain = None

try:
    from pip import main as foo #noqa
    if callable(foo):
        pipmain = foo
        logger.debug('found pipmain: from pip import main')
except:
    pass

try:
    from pip._internal import main as bar #noqa
    if callable(bar):
        pipmain = bar
        logger.debug('found pipmain: from pip._internal import main')
except:
    pass

try:
    from pip._internal.main import main as baz #noqa
    if callable(baz):
        pipmain = baz
        logger.debug('found pipmain: from pip._internal.main import main')
except:
    pass



from lambdatool.lambda_creator import LambdaCreator #noqa
from lambdatool.lambda_deployer import LambdaDeployer #noqa
from datetime import datetime

__title__ = 'lambdatool'
__version__ = '0.9.3'
__author__ = 'Mr. Chuck Muckamuck'
__author_email__ = 'chuck.muckamuck@gmail.com'
__license__ = '''MIT License

Copyright (c) 2017 Chuck Muckamuck

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
    SOFTWARE.
'''
__copyright__ = '(c) {}'.format(datetime.now().year)
__url__ = 'https://github.com/muckamuck/stackility'
__description__ = '''Sed sit amet congue augue. Phasellus aliquam leo non orci
egestas, in luctus enim lacinia. Praesent vulputate enim at placerat pharetra.
Aenean ullamcorper accumsan eros, ac porttitor nulla eleifend sed. In nunc
quam, fringilla sit amet sollicitudin eget, varius vel est. Phasellus mollis
libero eget sapien rhoncus, semper ultrices leo imperdiet. Suspendisse interdum
finibus turpis, id consequat quam pellentesque volutpat. Aenean diam ex,
tincidunt et lorem non, bibendum consequat nunc.
'''



class StackTool(object):
    _cf_client = None
    _stack_name = None
    _stage = None
    _region = None

    def __init__(self, stack_name, stage, profile, region, cf_client):
        """
        StackTool is a simple tool to print some specific data about a
        CloudFormation stack.

        Args:
            stack_name - name of the stack of interest
            stage - the supplied stage/environment
            profile - AWS credential profile (may be None)
            region - AWS region where the stack was created

        Returns:
           not a damn thing

        Raises:
            SystemError - if everything isn't just right
        """
        try:
            self._stack_name = stack_name
            self._stage = stage
            self._region = region
            self._cf_client = cf_client
        except Exception:
            raise SystemError

    def print_stack_info(self):
        '''
        List resources from the given stack

        Args:
            None

        Returns:
            A dictionary filled resources or None if things went sideways
        '''
        try:
            rest_api_id = None
            deployment_found = False

            response = self._cf_client.describe_stack_resources(
                StackName=self._stack_name
            )

            print('\nThe following resources were created:')
            for resource in response['StackResources']:
                if resource['ResourceType'] == 'AWS::ApiGateway::RestApi':
                    rest_api_id = resource['PhysicalResourceId']
                elif resource['ResourceType'] == 'AWS::ApiGateway::Deployment':
                    deployment_found = True

                print('\t{}\t{}\t{}'.format(
                        resource['ResourceType'],
                        resource['LogicalResourceId'],
                        resource['PhysicalResourceId']
                    )
                )

            if rest_api_id and deployment_found:
                url = 'https://{}.execute-api.{}.amazonaws.com/{}'.format(
                    rest_api_id,
                    self._region,
                    self._stage
                )
                print('\nThe deployed service can be found at this URL:')
                print('\t{}\n'.format(url))

            return response
        except Exception as wtf:
            print(wtf)
            return None


if __name__ == '__main__':
    stack_tool = StackTool('lambda-mars-dev', 'dev', None, 'us-east-2')
    stack_tool.print_stack_info()



import boto3
import sys
import logging


def get_api_client(profile, region, aws_service):
    try:
        if profile and region:
            api_session = boto3.session.Session(profile_name=profile, region_name=region)
        elif profile:
            api_session = boto3.session.Session(profile_name=profile)
        elif region:
            api_session = boto3.session.Session(region_name=region)
        else:
            api_session = boto3.session.Session()

        api_client = api_session.client(aws_service)
        return api_client
    except Exception as x:
        logging.fatal('Exception caught in get_api_client(): {}'.format(x))
        sys.exit(1)



'''
Stuff for the lambda role
'''
role_parameter_section = """  role:
    Description: name of the role
    Type: String"""
parameter_role_spec = 'Ref: role'
imported_role_spec = 'Fn::ImportValue: {}'

'''
Stuff for the subnet(s)
'''
subnets_parameter_section = """  subnetIds:
    Description: list of subnets
    Type: CommaDelimitedList"""
subnets_parameter_spec = 'Ref: subnetIds'
imported_subnets_spec = '- Fn::ImportValue: {}'

'''
Stuff for the security group(s)
'''
sg_parameter_section = """  securityGroupIds:
    Description: list of security groups
    Type: CommaDelimitedList"""
sg_parameter_spec = 'Ref: securityGroupIds'
imported_sg_spec = '- Fn::ImportValue: {}'

api_output_section = '''Outputs:
  LambdaFunction:
    Description: The name of a LambdaTool function
    Value:
      Ref: LambdaFunction
    Export:
      Name: {}-Name
  LambdaFunctionARN:
    Description: The ARN of a LambdaTool function
    Value:
      Fn::GetAtt:
        - LambdaFunction
        - Arn
    Export:
      Name: {}-Arn
  RestAPIid:
    Description: The ID of the resulting RestAPI
    Value:
      Ref: theAPI
    Export:
      Name: {}-RestAPI'''

output_section = '''Outputs:
  LambdaFunction:
    Description: The name of a LambdaTool function
    Value:
      Ref: LambdaFunction
    Export:
      Name: {}-Name
  LambdaFunctionARN:
    Description: The ARN of a LambdaTool function
    Value:
      Fn::GetAtt:
        - LambdaFunction
        - Arn
    Export:
      Name: {}-Arn'''

lambda_log_group = '''  LambdaLogGroup:
    Type: AWS::Logs::LogGroup
    Properties:
      LogGroupName:
        Ref: logGroupName
      RetentionInDays:
        Ref: retentionDays'''

resource_policy_spacer = '                  - '
resource_policy = '''
      Policy:
        Version: '2012-10-17'
        Statement:
          - Effect: Allow
            Principal: '*'
            Action: execute-api:Invoke
            Resource: arn:aws:execute-api:*:*:*
            Condition:
              IpAddress:
                aws:SourceIp:
'''



import logging
import traceback
import subprocess
import os
import sys
import shutil
import lambdatool.utility
import json

logger = logging.getLogger()
logger.setLevel(logging.INFO)
logging.basicConfig(level=logging.INFO,
                    format='[%(levelname)s] %(asctime)s (%(module)s) %(message)s',
                    datefmt='%Y/%m/%d-%H:%M:%S')

IGNORED_STUFF = ('template_template', '*.pyc')


class LambdaCreator:
    """
    Lambda utility is yet another tool create and deploy AWS lambda functions
    """
    def __init__(self, config_block):
        """
        Lambda utility init method.

        Args:
            config_block - a dictionary created in the CLI driver. See that
                           script for the things that are required and
                           optional.

        Returns:
           not a damn thing

        Raises:
            SystemError - if everything isn't just right
        """
        if config_block:
            self._config = config_block
            self._profile = config_block['profile']
            self._region = config_block['region']
            self._service = config_block['service']
        else:
            logger.error('config block was garbage')
            raise SystemError

    def create_lambda(self):
        """
        Write the template for a new lambda to the indicated target directory.

        Args:
            None

        Returns:
            True if the lambda is created
            False if the lambda is not created for some odd reason
        """
        try:
            destination_directory = '{}/{}'.format(
                self._config['directory'],
                self._config['name']
            )

            if os.path.exists(destination_directory):
                print('{} already exists, exiting'.format(destination_directory))
                sys.exit(1)
            else:
                logger.info('     source_directory: {}'.format(self._config['template_directory']))
                logger.info('destination_directory: {}'.format(destination_directory))

            default_vpc_info = self._describe_lambda_environment()
            logger.debug(json.dumps(self._config, indent=2))
            logger.debug(json.dumps(default_vpc_info, indent=2))

            shutil.copytree(
                self._config['template_directory'],
                destination_directory,
                symlinks=False,
                ignore=shutil.ignore_patterns(*IGNORED_STUFF)
            )

            self.write_config_ini(destination_directory, default_vpc_info)

            dot_lambdatool = '{}/.lambdatool'.format(destination_directory)
            meta_data = None
            with open(dot_lambdatool, 'r') as f:
                meta_data = json.load(f)
                meta_data['name'] = self._config['name']

            with open(dot_lambdatool, 'w') as f:
                json.dump(meta_data, f, indent=4)

            return True
        except Exception as x:
            logger.error('Exception caught in create_lambda(): {}'.format(x))
            traceback.print_exc(file=sys.stdout)
            return False

    def _describe_lambda_environment(self):
        '''
        Find the default vpc for the given region

        Args:
            None

        Returns:
            a dictionary the contains the  vpc ID, list subnets and default
            security group
        '''
        try:
            vpc_info = {}
            ec2_client = lambdatool.utility.get_api_client(
                self._profile,
                self._region,
                'ec2'
            )

            response = ec2_client.describe_vpcs()

            for vpc in response['Vpcs']:
                if vpc['IsDefault']:
                    subnets = self._find_default_subnets(vpc['VpcId'])
                    if subnets:
                        vpc_info['subnets'] = subnets

                    security_group = self._find_default_security_group(vpc['VpcId'])
                    if security_group:
                        vpc_info['security_group'] = security_group

                    lambda_role = self._find_lambda_role()
                    if lambda_role:
                        vpc_info['role'] = lambda_role

                    logger.info(json.dumps(vpc_info, indent=2))
                    return vpc_info
        except Exception as wtf:
            logger.error('Exception caught in create_lambda(): {}'.format(wtf))
            logger.fatal(wtf, exc_info=False)
            sys.exit(1)

        return {}

    def _find_lambda_role(self):
        try:
            iam_client = lambdatool.utility.get_api_client(
                self._profile,
                self._region,
                'iam'
            )

            response = iam_client.list_roles(MaxItems=5)
            while response:
                for role in response['Roles']:
                    if role['RoleName'] == 'lambda_basic_vpc_execution':
                        logger.info('found role: {}'.format(role['Arn']))
                        return role['Arn']

                if response['IsTruncated']:
                    response = iam_client.list_roles(
                        MaxItems=5,
                        Marker=response['Marker']
                    )
                else:
                    response = None
        except Exception as wtf:
            logger.error('Exception caught in create_lambda(): {}'.format(wtf))
            traceback.print_exc(file=sys.stdout)

        return None

    def _find_default_security_group(self, vpc_id):
        try:
            ec2_client = lambdatool.utility.get_api_client(
                self._profile,
                self._region,
                'ec2'
            )

            response = ec2_client.describe_security_groups(
                Filters=[{'Name': 'vpc-id', 'Values': [vpc_id]}]
            )

            for sg in response['SecurityGroups']:
                logger.info('found candidate security group: {}'.format(sg['GroupId']))
                if sg['GroupName'] == 'default':
                    logger.info('found security group: {}'.format(sg['GroupId']))
                    return sg['GroupId']
        except Exception as wtf:
            logger.error('Exception caught in create_lambda(): {}'.format(wtf))
            traceback.print_exc(file=sys.stdout)

        return None

    def _find_default_subnets(self, vpc_id):
        try:
            subnets = []
            ec2_client = lambdatool.utility.get_api_client(
                self._profile,
                self._region,
                'ec2'
            )

            response = ec2_client.describe_subnets(
                Filters=[{'Name': 'vpc-id', 'Values': [vpc_id]}]
            )

            for subnet in response['Subnets']:
                if subnet['DefaultForAz'] == True:
                    subnets.append(subnet['SubnetId'])

            logger.info('Found subnets: {}'.format(subnets))
            return subnets
        except Exception as wtf:
            logger.error('Exception caught in create_lambda(): {}'.format(wtf))
            traceback.print_exc(file=sys.stdout)

        return None

    def deploy_lambda(self):
        """
        Deploy an existing lambda to the indicated by creating CF template...

        Args:
            None

        Returns:
            True if the lambda is deployed
            False if the lambda is not deployed for some odd reason
        """
        try:
            logger.info(self._config)
            cwd = os.getcwd()
            dirs = cwd.split('/')
            lambda_name = dirs[-1]
            logger.info('lambda_name: {}'.format(lambda_name))
            return True
        except Exception as x:
            logger.error('Exception caught in deploy_lambda(): {}'.format(x))
            traceback.print_exc(file=sys.stdout)
            return False

    def execute_command(self, command):
        buf = ""
        try:
            p = subprocess.Popen(command, stdout=subprocess.PIPE)
            out, err = p.communicate()
            for c in out:
                buf = buf + c
            return p.returncode, buf
        except subprocess.CalledProcessError as x:
            logger.error('Exception caught in create_lambda(): {}'.format(x))
            traceback.print_exc(file=sys.stdout)
            return False
            return x.returncode, None

    def write_config_ini(self, destination_directory, env_info):
        file_name = '{}/config/config.ini'.format(destination_directory)
        with open(file_name, 'w') as ini_file:
            ini_file.write('[dev]\n')
            if 'security_group' in env_info:
                ini_file.write('security_group={}\n'.format(env_info['security_group']))
            else:
                ini_file.write('#security_group=ENTER_A_SECURITY_GROUP_FOR_VPC\n')

            if 'subnets' in env_info:
                wrk = str()
                for subnet in env_info['subnets']:
                    if len(wrk) == 0:
                        wrk = subnet
                    else:
                        wrk = '{},{}'.format(wrk, subnet)

                ini_file.write('subnets={}\n'.format(wrk))
            else:
                ini_file.write('#subnets=ENTER_A_COMMA_SEPARATED_LIST_OF_VPC_SUBNETS\n')

            if 'role' in env_info:
                ini_file.write('role={}\n'.format(env_info['role']))
            else:
                role_arn = input("Enter execution role ARN (smash enter to skip): ")
                parts = role_arn.split(':')
                if len(parts) == 6 and role_arn.startswith('arn:aws:iam'):
                    ini_file.write('role={}\n'.format(role_arn))
                else:
                    ini_file.write('role=ADD_YOUR_LAMBDA_IAM_ROLE\n')

            ini_file.write('memory=512\n')

            if self._service:
                ini_file.write('service=true\n')
            else:
                ini_file.write('service=false\n')

            bucket_name = input("Enter artifact bucket (smash enter to skip): ")
            if len(bucket_name) == 0:
                ini_file.write('bucket=ADD_YOUR_ARTIFACT_BUCKET\n')
            else:
                ini_file.write('bucket={}\n'.format(bucket_name))

            ini_file.write('\n')



"""
The command line interface to stackility.

Major help from: https://www.youtube.com/watch?v=kNke39OZ2k0
"""
import lambdatool
from lambdatool.lambda_creator import LambdaCreator
from lambdatool.lambda_deployer import LambdaDeployer
import click
import boto3
import logging
import sys
import os
import json

default_stage = 'dev'
fresh_notes = '''A skeleton of the new lambda, {}, has been created.

In {}/{}/config you will find a config.ini file that you should
fill in with parameters for your own account.

Develop the lambda function as needed then you can deploy it with:
lambdatool deploy. The lambda has been started in main.py.
'''


@click.group()
@click.version_option(version='0.9.3')
def cli():
    pass


@cli.command()
@click.option('-d', '--directory', help='target directory for new Lambda, defaults to current directory')
@click.option('-n', '--name', help='name of the new lambda skeleton', required=True)
@click.option('-s', '--service', help='create a flask like micro-service', is_flag=True)
@click.option('-p', '--profile', help='AWS CLI profile to use in the deployment, more details at http://docs.aws.amazon.com/cli/latest/userguide/cli-chap-getting-started.html')
@click.option('-r', '--region', help='target region, defaults to your credentials default region')
def new(directory, name, service, profile, region):
    command_line = {}
    command_line['name'] = name

    if service:
        command_line['template_directory'] = '{}/template/service'.format(lambdatool.__path__[0])
    else:
        command_line['template_directory'] = '{}/template/simple'.format(lambdatool.__path__[0])

    if directory:
        command_line['directory'] = directory
    else:
        command_line['directory'] = '.'

    if profile:
        command_line['profile'] = profile
    else:
        command_line['profile'] = None

    if region:
        command_line['region'] = region
    else:
        command_line['region'] = None

    command_line['service'] = service

    if start_new_lambda(command_line):
        sys.exit(0)
    else:
        sys.exit(1)


@cli.command()
@click.option('-d', '--directory', help='scratch directory for deploy, defaults to /tmp')
@click.option('-s', '--stage', help='environment/stage used to name and deploy the Lambda function, defaults to dev')
@click.option('-p', '--profile', help='AWS CLI profile to use in the deployment, more details at http://docs.aws.amazon.com/cli/latest/userguide/cli-chap-getting-started.html')
@click.option('-r', '--region', help='target region, defaults to your credentials default region')
def deploy(directory, stage, profile, region):
    command_line = {}

    if directory:
        command_line['work_directory'] = directory
    else:
        command_line['work_directory'] = '/tmp'

    if stage:
        command_line['stage'] = stage
    else:
        command_line['stage'] = default_stage

    if profile:
        command_line['profile'] = profile
    else:
        command_line['profile'] = None

    if region:
        command_line['region'] = region
    else:
        command_line['region'] = None

    command_line['template_directory'] = '{}/template'.format(lambdatool.__path__[0])
    logging.info('command_line: {}'.format(json.dumps(command_line, indent=2)))

    if deploy_lambda(command_line):
        sys.exit(0)
    else:
        sys.exit(1)


@cli.command()
@click.option('-s', '--stage', help='environment/stage of interest', required=True)
def print_env(stage):
    config_file = f'config/{stage}/function.properties'
    if os.path.isfile(config_file):
        with open(config_file, 'r') as f:
            tmp = f.readline()
            while tmp:
                food = tmp.strip()
                print(f'export {food}')
                tmp = f.readline()



def start_new_lambda(command_line):
    try:
        tool = LambdaCreator(command_line)
    except Exception:
        sys.exit(1)

    if tool.create_lambda():
        logging.info('create_new_lambda() went well')
        print('\n\n\n\n')
        print('********************************************************************************')
        print(fresh_notes.format(
            command_line['name'],
            command_line['directory'],
            command_line['name'])
        )
    else:
        logging.error('create_new_lambda() did not go well')
        sys.exit(1)


def deploy_lambda(command_line):
    try:
        tool = LambdaDeployer(command_line)
    except Exception:
        sys.exit(1)

    if tool.deploy_lambda():
        logging.info('deploy_lambda() went well')
        return True
    else:
        logging.error('deploy_lambda() did not go well')
        sys.exit(1)


def find_myself():
    s = boto3.session.Session()
    return s.region_name



region = None
stage = None
account = None
lambda_name = None

the_api = """  theAPI:
    Type: AWS::ApiGateway::RestApi
    DependsOn: LambdaFunction
    Properties:
      Description: LambdaTool created this AWS ApiGateway RestApi thing
      {resource_policy}
      Body:
        swagger: "2.0"
        info:
          version: "2017-11-15T16:30:51Z"
          title: "{short_name}-{stage_name}"
        host: "ozi3yy5k9a.execute-api.{region}.amazonaws.com"
        basePath: "/{stage_name}"
        schemes:
        - "https"
        paths:
          /:
            x-amazon-apigateway-any-method:
              produces:
              - "application/json"
              responses:
                '200':
                  description: "200 response"
                  schema:
                    $ref: "#/definitions/Empty"
              x-amazon-apigateway-integration:
                responses:
                  default:
                    statusCode: "200"
                uri: "arn:aws:apigateway:{region}:lambda:path/2015-03-31/functions/arn:aws:lambda:{region}:{account}:function:{short_name}-{stage_name}/invocations"
                passthroughBehavior: "when_no_match"
                httpMethod: "POST"
                contentHandling: "CONVERT_TO_TEXT"
                type: "aws_proxy"
          /{{proxy+}}:
            options:
              consumes:
              - "application/json"
              produces:
              - "application/json"
              responses:
                '200':
                  description: "200 response"
                  schema:
                    $ref: "#/definitions/Empty"
                  headers:
                    Access-Control-Allow-Origin:
                      type: "string"
                    Access-Control-Allow-Methods:
                      type: "string"
                    Access-Control-Allow-Headers:
                      type: "string"
              x-amazon-apigateway-integration:
                responses:
                  default:
                    statusCode: "200"
                    responseParameters:
                      method.response.header.Access-Control-Allow-Methods: "'DELETE,GET,HEAD,OPTIONS,PATCH,POST,PUT'"
                      method.response.header.Access-Control-Allow-Headers: "'Content-Type,Authorization,X-Amz-Date,X-Api-Key,X-Amz-Security-Token'"
                      method.response.header.Access-Control-Allow-Origin: "'*'"
                requestTemplates:
                  application/json: "{{\\"statusCode\\": 200}}"
                passthroughBehavior: "when_no_match"
                type: "mock"
            x-amazon-apigateway-any-method:
              produces:
              - "application/json"
              parameters:
              - name: "proxy"
                in: "path"
                required: true
                type: "string"
              responses: {{}}
              x-amazon-apigateway-integration:
                responses:
                  default:
                    statusCode: "200"
                uri: "arn:aws:apigateway:{region}:lambda:path/2015-03-31/functions/arn:aws:lambda:{region}:{account}:function:{short_name}-{stage_name}/invocations"
                passthroughBehavior: "when_no_match"
                httpMethod: "POST"
                cacheNamespace: "fyc8uq"
                cacheKeyParameters:
                - "method.request.path.proxy"
                contentHandling: "CONVERT_TO_TEXT"
                type: "aws_proxy"
        definitions:
          Empty:
            type: "object"
            title: "Empty Schema"
        x-amazon-apigateway-binary-media-types:
        - "*/*"
  theDeployment:
    DependsOn: theAPI
    Properties:
      Description:
        "{stage_name}"
      RestApiId:
        Ref: theAPI
      StageName:
        "{stage_name}"
    Type: AWS::ApiGateway::Deployment
  APIGPermission:
    Type: AWS::Lambda::Permission
    DependsOn: theAPI
    Properties:
      FunctionName:
        Fn::GetAtt: [LambdaFunction, Arn]
      Action: lambda:InvokeFunction
      Principal: apigateway.amazonaws.com"""


def get_the_api_chunk(**kwargs):
    return the_api.format(
        region=kwargs['region'],
        stage_name=kwargs['stage_name'],
        short_name=kwargs['short_name'],
        account=kwargs['account'],
        resource_policy=kwargs['resource_policy']
    )

if __name__ == '__main__':
    print(get_the_api_chunk(
            region='us-north-42',
            stage_name='exp',
            short_name='fred',
            account='0187378482993'
        )
    )



from mako.template import Template
from mako.runtime import Context

try:
    from StringIO import StringIO
except ImportError:
    from io import StringIO

from lambdatool.parts import get_the_api_chunk

from lambdatool.cf_import_things import role_parameter_section
from lambdatool.cf_import_things import parameter_role_spec
from lambdatool.cf_import_things import imported_role_spec

from lambdatool.cf_import_things import sg_parameter_section
from lambdatool.cf_import_things import sg_parameter_spec
from lambdatool.cf_import_things import imported_sg_spec

from lambdatool.cf_import_things import subnets_parameter_section
from lambdatool.cf_import_things import subnets_parameter_spec
from lambdatool.cf_import_things import imported_subnets_spec
from lambdatool.cf_import_things import api_output_section
from lambdatool.cf_import_things import output_section
from lambdatool.cf_import_things import lambda_log_group

from lambdatool.cf_import_things import resource_policy_spacer
from lambdatool.cf_import_things import resource_policy

import traceback
import os
import sys
import logging


whitelist = 'whitelist'
export_name = 'export_name'
snsTopicARN = 'snstopicarn'
trustedService = 'trustedservice'
trustedAccount = 'trustedaccount'
reservedConcurrency = 'reservedconcurrency'
schedule = 'scheduleexpression'
service = 'service'
new_line = '\n'
spacer = '          '

sns_topic_arn = """  snsTopicARN:
    Description: the ARN of the topic to which we are subscribing
    Type: String"""

trusted_service = """  trustedService:
    Description: service which this lambda trusts
    Type: String"""

trusted_account = """  trustedAccount:
    Description: account which this lambda trusts
    Type: String"""

reserved_concurrency = """  reservedConcurrency:
    Description: the maximum number of concurrent invocations
    Type: String"""

schedule_expression = """  scheduleExpression:
    Description: rate or cron expression for a scheduled  lambda
    Type: String"""

sns_subcription_resource = """  TopicSubscription:
    Type: AWS::SNS::Subscription
    DependsOn: LambdaFunction
    Properties:
      Endpoint:
        Fn::GetAtt: [LambdaFunction, Arn]
      Protocol: lambda
      TopicArn:
        Ref: snsTopicARN
  TopicPermission:
    Type: AWS::Lambda::Permission
    DependsOn: TopicSubscription
    Properties:
      FunctionName:
        Fn::GetAtt: [LambdaFunction, Arn]
      Action: lambda:InvokeFunction
      Principal: sns.amazonaws.com"""

WITH_ACCOUNT = 'WITH_ACCOUNT'
WITHOUT_ACCOUNT = 'WITHOUT_ACCOUNT'

trusted_service_resource = dict()
trusted_service_resource[WITHOUT_ACCOUNT] = """  TrustedService:
    Type: AWS::Lambda::Permission
    DependsOn: LambdaFunction
    Properties:
      FunctionName:
        Fn::GetAtt: [LambdaFunction, Arn]
      Action: lambda:InvokeFunction
      Principal:
        Ref: trustedService"""

trusted_service_resource[WITH_ACCOUNT] = """  TrustedService:
    Type: AWS::Lambda::Permission
    DependsOn: LambdaFunction
    Properties:
      FunctionName:
        Fn::GetAtt: [LambdaFunction, Arn]
      Action: lambda:InvokeFunction
      Principal:
        Ref: trustedService
      SourceAccount:
        Ref: trustedAccount"""

reserved_concurrency_resource = '''      ReservedConcurrentExecutions:
        Ref: reservedConcurrency'''

rule_id = '{}-{}'.format(
    os.environ.get('LAMBDA_NAME', 'unknown'),
    os.environ.get('ENVIRONMENT', 'none')
)

schedule_resource = """  LambdaSchedule:
    Type: AWS::Events::Rule
    DependsOn: LambdaFunction
    Properties:
      Description: String
      ScheduleExpression:
          Ref: scheduleExpression
      State: ENABLED
      Targets:
        -
          Arn:
            Fn::GetAtt: [LambdaFunction, Arn]
          Id:
            {}
  EventPermission:
    Type: AWS::Lambda::Permission
    DependsOn: LambdaFunction
    Properties:
      FunctionName:
        Fn::GetAtt: [LambdaFunction, Arn]
      Action: lambda:InvokeFunction
      Principal: events.amazonaws.com""".format(rule_id)


class TemplateCreator:
    _stack_properties = None
    _input_file = None
    _output_file = None
    _template_file = None
    _sns_topic_arn_found = False
    _trusted_service_found = False
    _trusted_account_found = False
    _reserved_concurrency_found = False
    _create_service = False
    _schedule_found = False
    _export_name = None
    _region = None
    _stage_name = None
    _short_name = None
    _account = None
    _ssm_client = None
    _import_role = False
    _import_subnets = False
    _import_security_group = False
    _import_sqs_arn = False
    _description = None
    _create_log_group = False
    SSM = '[ssm:'
    IMPORT = '[import:'

    _food = """      Environment:
        Variables:
"""

    def __init__(self, ssm_client):
        self._ssm_client = ssm_client

    def _prop_to_yaml(self, thing):
        idx = thing.find('=')
        if idx > -1:
            key = thing[:idx]
            val = thing[(idx+1):].strip()
            val = self._get_ssm_parameter(val)
            if val:
                return key, val

        return None, None

    def _inject_stuff(self):
        try:
            with open(self._input_file, 'r') as infile:
                for thing in infile:
                    key, val = self._prop_to_yaml(thing.strip())
                    if key and val:
                        self._food += spacer + key + ': ' + val + '\n'

            buf = StringIO()
            t = Template(filename=self._template_file)

            if self._sns_topic_arn_found:
                sns_var_bits = sns_topic_arn
                sns_resource_bits = sns_subcription_resource
            else:
                sns_var_bits = ''
                sns_resource_bits = ''

            if self._trusted_service_found:
                if self._trusted_account_found:
                    trusted_service_var_bits = trusted_service + new_line + trusted_account
                    trusted_service_resource_bits = trusted_service_resource[WITH_ACCOUNT]
                else:
                    trusted_service_var_bits = trusted_service
                    trusted_service_resource_bits = trusted_service_resource[WITHOUT_ACCOUNT]
            else:
                trusted_service_var_bits = ''
                trusted_service_resource_bits = ''

            if self._reserved_concurrency_found:
                reserved_concurrency_var_bits = reserved_concurrency
                reserved_concurrency_resource_bits = reserved_concurrency_resource
            else:
                reserved_concurrency_var_bits = ''
                reserved_concurrency_resource_bits = ''

            if self._schedule_found:
                schedule_var_bits = schedule_expression
                schedule_resource_bits = schedule_resource
            else:
                schedule_var_bits = ''
                schedule_resource_bits = ''

            if self._create_service:
                the_api_bits = get_the_api_chunk(
                    region=self._region,
                    stage_name=self._stage_name,
                    short_name=self._short_name,
                    account=self._account,
                    resource_policy=self._resource_policy
                )
            else:
                the_api_bits = ''

            if self._import_role:
                current_role_parameter_section = ''
                role = self._find_imported_csv(
                    self._stack_properties.get('role', None)
                )
                role_specification = imported_role_spec.format(role)
            else:
                current_role_parameter_section = role_parameter_section
                role_specification = parameter_role_spec

            subnet_specification = None
            includeVpcConfig = True
            if self._stack_properties.get('subnetIds', None) is None:
                current_subnets_parameter_section = None
                subnet_specification = None
                includeVpcConfig = False
            elif self._import_subnets:
                current_subnets_parameter_section = ''
                subnets = self._find_imported_csv(
                    self._stack_properties.get('subnetIds', None)
                )
                for subnet in subnets.split(','):
                    if subnet_specification:
                        subnet_specification = subnet_specification + \
                            '\n' + spacer + \
                            imported_subnets_spec.format(subnet)
                    else:
                        subnet_specification = imported_subnets_spec.format(subnet)
            else:
                current_subnets_parameter_section = subnets_parameter_section
                subnet_specification = subnets_parameter_spec

            sg_specification = None
            if self._stack_properties.get('securityGroupIds', None) is None:
                current_sg_parameter_section = None
                sg_specification = None
                includeVpcConfig = False
            elif self._import_security_group:
                current_sg_parameter_section = ''
                sg_csv = self._find_imported_csv(
                    self._stack_properties.get('securityGroupIds', None)
                )
                for sg in sg_csv.split(','):
                    if sg_specification:
                        sg_specification = sg_specification + \
                            '\n' + spacer + \
                            imported_sg_spec.format(sg)
                    else:
                        sg_specification = imported_sg_spec.format(sg)
            else:
                current_sg_parameter_section = sg_parameter_section
                sg_specification = sg_parameter_spec

            logging.info('includeVpcConfig: %s', includeVpcConfig)

            if self._export_name:
                if self._create_service:
                    output_section_bits = api_output_section.format(
                        self._export_name,
                        self._export_name,
                        self._export_name
                    )
                else:
                    output_section_bits = output_section.format(self._export_name, self._export_name)
            else:
                output_section_bits = ''

            if self._create_log_group:
                lambda_log_group_bits = lambda_log_group
            else:
                lambda_log_group_bits = ''

            ctx = Context(
                buf,
                environment_section=self._food,
                stackDescription=self._description,
                outputSection=output_section_bits,
                snsTopicARN=sns_var_bits,
                snsSubscriptionResource=sns_resource_bits,
                trustedService=trusted_service_var_bits,
                trustedServiceResource=trusted_service_resource_bits,
                reservedConcurrency=reserved_concurrency_var_bits,
                reservedConcurrencyResource=reserved_concurrency_resource_bits,
                scheduleExpression=schedule_var_bits,
                scheduleResource=schedule_resource_bits,
                theAPI=the_api_bits,
                roleParameterSection=current_role_parameter_section,
                roleSpecification=role_specification,
                subnetsParameterSection=current_subnets_parameter_section,
                subnetIds=subnet_specification,
                sgParameterSection=current_sg_parameter_section,
                lambdaLogGroup=lambda_log_group_bits,
                securityGroupIds=sg_specification,
                includeVpcConfig=includeVpcConfig
            )

            t.render_context(ctx)
            logging.info('writing template {}'.format(self._output_file))
            with open(self._output_file, "w") as outfile:
                    outfile.write(buf.getvalue())
        except Exception as wtf:
            logging.error('Exception caught in inject_stuff(): {}'.format(wtf))
            traceback.print_exc(file=sys.stdout)
            sys.exit(1)

    def _read_stack_properties(self):
        try:
            lowered_stack_properties = {}
            for key in self._stack_properties:
                lowered_key = key.lower()
                lowered_stack_properties[lowered_key] = self._stack_properties[key]

            role = lowered_stack_properties.get('role', None)
            subnets = lowered_stack_properties.get('subnetids', None)
            security_group = lowered_stack_properties.get('securitygroupids', None)
            sqs_arn = lowered_stack_properties.get('sqsarn', None)

            if role and role.startswith(self.IMPORT):
                self._import_role = True

            if subnets and subnets.startswith(self.IMPORT):
                self._import_subnets = True

            if security_group and security_group.startswith(self.IMPORT):
                self._import_security_group = True

            if sqs_arn and sqs_arn.startswith(self.IMPORT):
                self._import_sqs_arn = True

            if snsTopicARN in lowered_stack_properties:
                self._sns_topic_arn_found = True

            if trustedService in lowered_stack_properties:
                self._trusted_service_found = True

            if trustedAccount in lowered_stack_properties:
                self._trusted_account_found = True

            if reservedConcurrency in lowered_stack_properties:
                self._reserved_concurrency_found = True

            if schedule in lowered_stack_properties:
                self._schedule_found = True

            if export_name in lowered_stack_properties:
                self._export_name = lowered_stack_properties.get(export_name)

            if lowered_stack_properties.get('whitelist', None):
                parts = lowered_stack_properties.get(whitelist).split(',')
                buf = resource_policy

                for cidr in parts:
                    buf = buf + '\n' + resource_policy_spacer + cidr

                self._resource_policy = buf
            else:
                self._resource_policy = ''

            tmp = lowered_stack_properties.get(service, 'false').lower()
            if tmp == 'true':
                self._create_service = True
        except Exception as wtf:
            logging.error('Exception caught in read_stack_properties(): {}'.format(wtf))
            sys.exit(1)

        return True

    def create_template(self, **kwargs):
        try:
            self._input_file = kwargs['function_properties']
            self._stack_properties = kwargs['stack_properties']
            self._output_file = kwargs['output_file']
            self._template_file = kwargs['template_file']
            self._region = kwargs['region']
            self._stage_name = kwargs['stage_name']
            self._short_name = kwargs['short_name']
            self._account = kwargs['account']
            self._create_log_group = kwargs['create_log_group']
            self._description = kwargs.get('description', 'Fantastic Lambda Function')

            self._read_stack_properties()
            self._inject_stuff()
            return True
        except Exception as wtf:
            logging.error(wtf)
            return False

    def _get_ssm_parameter(self, p):
        """
        Get parameters from Simple Systems Manager

        Args:
            p - a parameter name

        Returns:
            a value, decrypted if needed, if successful or None if things go
            sideways.
        """
        try:
            if p.startswith(self.SSM) and p.endswith(']'):
                parts = p.split(':')
                p = parts[1].replace(']', '')
            else:
                return p

            response = self._ssm_client.get_parameter(Name=p, WithDecryption=True)
            return response.get('Parameter', {}).get('Value', None)
        except Exception as ruh_roh:
            logging.error(ruh_roh, exc_info=False)

        return None

    def _find_imported_csv(self, raw_str):
        answer = None
        try:
            wrk = raw_str
            wrk = wrk.replace('[', '')
            wrk = wrk.replace(']', '')
            wrk = wrk.replace(' ', '')
            parts = wrk.split(':')
            answer = parts[1]
        except Exception:
            answer = None

        return answer


if __name__ == '__main__':
    templateCreator = TemplateCreator()
    templateCreator.create_template(
        function_properties='/tmp/scratch/f315ee80/config/dev/function.properties',
        stack_properties='/tmp/scratch/f315ee80/stack.properties',
        output_file='/tmp/template.yaml',
        template_file='template_template'
    )



from setuptools import setup
import os


def find_data(starting_dir, the_dir):
    original_cwd = os.getcwd()
    tree = []
    try:
        os.chdir(starting_dir)
        for folder, subs, files in os.walk(the_dir):
            for file in files:
                tree.append('{}/{}'.format(folder, file))
    except Exception:
        pass

    os.chdir(original_cwd)
    return tree


setup(
    name='LambdaTool',
    version='0.9.3',
    packages=['lambdatool'],
    description='Python Lambda utility',
    author='Chuck Muckamuck',
    author_email='Chuck.Muckamuck@gmail.com',
    include_package_data=True,
    package_data={'lambdatool': find_data('lambdatool', 'template')},
    install_requires=[
        'boto3>=1.4.3',
        'GitPython>=2.1.7',
        'Click>=6.7',
        'PyYAML>=3.12',
        'pymongo>=3.4.0',
        'stackility>=0.3',
        'Mako>=1.0.6'
    ],
    entry_points="""
        [console_scripts]
        lambdatool=lambdatool.command:cli
    """
)
